# FuzzyOcr plugin, version 2.2
# Changelog:
#    version 2.0
#	Replaced imagemagick with netpbm
#	Invoke giffix to fix broken gifs before conversion
#	Support png images
#	Analyze the file to detect the format without content-type
#	Added several configuration parameters
#    version 2.1
#       Added scoring for wrong content-type
#       Added scoring for broken gif images
#       Added configuration for helper applications
#       Added autodisable_score feature to disable the OCR engine if the message has already enough points
#    version 2.1b
#       Rule bugfix to avoid warnings
#    version 2.1c
#       Applied patch provided by Howard Kash to fix problems with spamassassin + Mailscanner + FuzzyOcr
#       Removed '-' from jpegtopnm arguments to provide backwards compatibility for older netpbm versions
#       Fixed typo (treshold -> threshold)
#    version 2.2
#       Small bugfix in content-type check for jpeg (jpg was not matching), thanks to Matthias Keller
#       Added more error handling
#       Removed debug files, added logfile instead
#       More messages with verbose = 2
#
# written by Christian Holler decoder_at_own-hero_dot_net

package FuzzyOcr;

use strict;
use Mail::SpamAssassin;
use Mail::SpamAssassin::Util;
use Mail::SpamAssassin::Plugin;

use String::Approx 'adistr';

our @ISA = qw (Mail::SpamAssassin::Plugin);

our @err_msges = (
"Failed to open pipe to external programs for %s image processing.
Please check that all helper programs are installed and in the correct path.
Also make sure that \"%s\" is writeable for the plugin.",
"Unexpected error in pipe to external programs.
Please check that all helper programs are installed and in the correct path.
Also make sure that %s is writeable for the plugin. (Pipe exit code %d, Image type %d)",
"Unexpected error while trying to reopen \"%s\" for reading.
Probably the file was deleted or moved while the plugin was running."
);

our @words = ( );
our $self;
our $pms;

# Default values
our $threshold = "0.3";
our $base_score = "4";
our $add_score = "1";
our $wctypescore = "1.5";
our $cimgscore = "3";
our $countreq = 2;
our $verbose = 1;
our $tmppath = "/tmp";
our $giffix = "/usr/bin/giffix";
our $giftopnm = "/usr/bin/giftopnm";
our $jpegtopnm = "/usr/bin/jpegtopnm";
our $pngtopnm = "/usr/bin/pngtopnm";
our $gocr = "/usr/bin/gocr";
our $dscore = 50;
our $logfile = "/etc/mail/spamassassin/FuzzyOcr.log";

# constructor: register the eval rule
sub new {
   my ( $class, $mailsa ) = @_;
   $class = ref($class) || $class;
   my $self = $class->SUPER::new($mailsa);
   bless( $self, $class );
   $self->register_eval_rule("check_fuzzy_ocr");
   $self->register_eval_rule("dummy_check");
   return $self;
}

sub parse_config {
  my ($self, $opts) = @_;
  if ($opts->{key} eq "focr_word") {
	push(@words, $opts->{value});
  } elsif ($opts->{key} eq "focr_threshold") {
	$threshold = $opts->{value};
  } elsif ($opts->{key} eq "focr_base_score") {
        $base_score = $opts->{value};
  } elsif ($opts->{key} eq "focr_add_score") {
        $add_score = $opts->{value};
  } elsif ($opts->{key} eq "focr_corrupt_score") {
        $cimgscore = $opts->{value};
  } elsif ($opts->{key} eq "focr_wrongctype_score") {
        $wctypescore = $opts->{value};
  } elsif ($opts->{key} eq "focr_counts_required") {
        $countreq = $opts->{value};
  } elsif ($opts->{key} eq "focr_verbose") {
	$verbose = $opts->{value};
  } elsif ($opts->{key} eq "focr_tmp_path") {
	$tmppath = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_giffix") {
        $giffix = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_giftopnm") {
        $giftopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_jpegtopnm") {
        $jpegtopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_pngtopnm") {
        $pngtopnm = $opts->{value};
  } elsif ($opts->{key} eq "focr_bin_gocr") {
        $gocr = $opts->{value};
  } elsif ($opts->{key} eq "focr_autodisable_score") {
          $dscore = $opts->{value};
  } elsif ($opts->{key} eq "focr_logfile") {
	  $logfile = $opts->{value};
  }
}

sub dummy_check {
	return 0;
}

sub handle_error {
	my ($err_msg, @var_vals) = @_;
	$err_msg = sprintf($err_msg, @var_vals);
	logfile($err_msg);
}

sub logfile {
	my $logtext = $_[0];
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
	my $time = sprintf("%4d-%02d-%02d %02d:%02d:%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
	$logtext =~ s/\n/\n                      /g;
	open LOGFILE, ">> $logfile";
	print LOGFILE "[$time] $logtext\n";
	close LOGFILE;
}

sub debuglog {
	if ($verbose > 1) {
		logfile("Debug mode: $_[0]");
	}
}

sub wrong_ctype {
   my ( $format, $ctype ) = @_;
   if ($wctypescore) {
         my $debuginfo = "";
         if ($verbose > 0) {
                $debuginfo = ("Image has format \"$format\" but content-type is \"$ctype\"");
                debuglog($debuginfo);
         }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR_WRONG_CTYPE"} = sprintf("%0.3f", $wctypescore);
         }
         $pms->_handle_hit("FUZZY_OCR_WRONG_CTYPE", $wctypescore, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR_WRONG_CTYPE} . "\n$debuginfo");
   }
}

sub corrupt_img {
   my ( $err ) = @_;
   if ($cimgscore) {
         my $debuginfo = "";
         if ($verbose > 0) {
		chomp($err);
                $debuginfo = ("Corrupt image: $err");
                debuglog($debuginfo);
         }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR_CORRUPT_IMG"} = sprintf("%0.3f", $cimgscore);
         }
	 $pms->_handle_hit("FUZZY_OCR_CORRUPT_IMG", $cimgscore, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR_CORRUPT_IMG} . "\n$debuginfo");
   }
}

sub check_fuzzy_ocr {
   ( $self, $pms ) = @_;
   my @found = ( );
   my $image_type = 0;
   my $cnt = 0;

   debuglog("Starting FuzzyOcr...");

   foreach my $p ( $pms->{msg}->find_parts("image") ) {
      my $cscore = $pms->get_score();
      if ($cscore > $dscore) {
	 debuglog("Scan canceled, message has already more than $dscore points.");
         return 0;
      }
      my ( $ctype, $boundary, $charset, $name ) =
        Mail::SpamAssassin::Util::parse_content_type(
         $p->get_header('content-type') );
	 if ($ctype =~ /image/i) {
	 debuglog("Analyzing file with content-type \"$ctype\"");
	 my $firstline = ($p->decode())[0];
	 my $tempfile = $tmppath . "/" . "spamassassin.$$.focr";
	 my $errfile = $tmppath . "/" . "spamassassin.$$.focr.err";
	 if ($firstline =~ /^\x47\x49\x46/) {
	        if ($ctype !~ /gif/i) {
			wrong_ctype("GIF", $ctype);
		}
		$image_type = 1;
		unless (open IMAGE_PROCESSOR, "|$giffix 2>$errfile | $giftopnm - |$gocr -i - > $tempfile") {
			handle_error($err_msges[0], ("GIF", $tmppath));
			return 0;
		}
	 } elsif ($firstline =~ /^\xff\xd8/) {
                if ($ctype !~ /jpe{0,1}g/i) {
                        wrong_ctype("JPEG", $ctype);
                }
		$image_type = 2;
		unless (open IMAGE_PROCESSOR, "|$jpegtopnm |$gocr -i - > $tempfile") {
			handle_error($err_msges[0], ("JPEG", $tmppath));
			return 0;
		}
	 } elsif ($firstline =~ /^\x89\x50\x4e\x47/) {
                if ($ctype !~ /png/i) {
                        wrong_ctype("PNG", $ctype);
                }
		$image_type = 3;
		unless (open IMAGE_PROCESSOR, "|$pngtopnm - |$gocr -i - > $tempfile") {
			handle_error($err_msges[0], ("PNG", $tmppath));
			return 0;
		}
	 } else {
		$image_type = 0;
		next;
	 }
         foreach $p ( $p->decode() ) {
		print IMAGE_PROCESSOR $p;
         }

	 debuglog("Recognized file type: $image_type");

	 close IMAGE_PROCESSOR;
	 if ($?) {
		my @var_vals = ($tmppath, $?, $image_type);
		handle_error($err_msges[1], ($tmppath, $?, $image_type));
		return 0;
	 }
	 if (-r $errfile) {
		open ERRFILE, "<$errfile";
		while (<ERRFILE>) {
			if ($_ =~ /GIF-LIB error/i) {
				corrupt_img($_);
				last;
			}
		}
		close ERRFILE;
		if ($verbose < 2) {
			unlink $errfile;
		} else {
			debuglog("File \"$errfile\" not unlinked, because debug mode is enabled.");
		}
	 }
	 unless (open OCR_DATA, "<$tempfile") {
		 handle_error($err_msges[2], ($tempfile));
		 return 0;
	 }
	 while (<OCR_DATA>) {
	    s/[^a-zA-Z ]//g;
	    $_ = lc;
            my $w;
            foreach $w (@words) {
                $w = lc $w;
                my $matched = adistr($w, $_);
                if (abs($matched) < $threshold) {
                        $cnt++;
			if ($verbose > 0) {
				push(@found, "\"$w\"" . " with fuzz of " . abs($matched));
			}
                }
            }
         }
	 close OCR_DATA;
	 if ($verbose < 2) {
		unlink $tempfile;
	 } else {
		debuglog("File \"$tempfile\" not unlinked, because debug mode is enabled.");
	 }
      }
   }
   if ($cnt >= $countreq) {
	 my $score = $base_score + ($cnt - $countreq) * $add_score;
	 my $debuginfo = "";
	 if ($verbose > 0) {
		$debuginfo = ("Words found:\n" . join("\n", @found) ."\n($cnt word occurrences found)");
		debuglog($debuginfo);
	 }
         for my $set (0..3) {
            $pms->{conf}->{scoreset}->[$set]->{"FUZZY_OCR"} = sprintf("%0.3f", $score);
         }
         $pms->_handle_hit("FUZZY_OCR", $score, "BODY: ", $pms->{conf}->{descriptions}->{FUZZY_OCR} . "\n$debuginfo");
   }
   debuglog("FuzzyOcr ending successfully...");
   return 0;
}

1;
